@@ -1,42 +1,43 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level
-#of this repository contains the full copyright notices and license terms.
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
 import datetime
+
 from decimal import Decimal, ROUND_HALF_EVEN
-from trytond.model import ModelView, ModelSQL, fields
-from trytond.tools import safe_eval, datetime_strftime
+
+from sql import Window
+from sql.functions import NthValue
+
+from trytond.model import (
+    ModelView, ModelSQL, DeactivableMixin, fields, Unique, Check)
+from trytond.tools import datetime_strftime
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 from trytond.rpc import RPC
+from trytond.pyson import Eval
 
 __all__ = ['Currency', 'Rate']
 
 
-class Currency(ModelSQL, ModelView):
+class Currency(DeactivableMixin, ModelSQL, ModelView):
     'Currency'
     __name__ = 'currency.currency'
-    name = fields.Char('Name', required=True, translate=True)
-    symbol = fields.Char('Symbol', size=10, required=True)
-    code = fields.Char('Code', size=3, required=True)
-    numeric_code = fields.Char('Numeric Code', size=3)
-    rate = fields.Function(fields.Numeric('Current rate', digits=(12, 6),
-        on_change_with=['rates']), 'get_rate')
-    rates = fields.One2Many('currency.currency.rate', 'currency', 'Rates')
-    rounding = fields.Numeric('Rounding factor', digits=(12, 6), required=True)
-    digits = fields.Integer('Display Digits', required=True)
-    active = fields.Boolean('Active')
-
-    # monetary formatting
-    mon_grouping = fields.Char('Grouping', required=True)
-    mon_decimal_point = fields.Char('Decimal Separator', required=True)
-    mon_thousands_sep = fields.Char('Thousands Separator')
-    p_sign_posn = fields.Integer('Positive Sign Position', required=True)
-    n_sign_posn = fields.Integer('Negative Sign Position', required=True)
-    positive_sign = fields.Char('Positive Sign')
-    negative_sign = fields.Char('Negative Sign')
-    p_cs_precedes = fields.Boolean('Positive Currency Symbol Precedes')
-    n_cs_precedes = fields.Boolean('Negative Currency Symbol Precedes')
-    p_sep_by_space = fields.Boolean('Positive Separate by Space')
-    n_sep_by_space = fields.Boolean('Negative Separate by Space')
+    name = fields.Char('Name', required=True, translate=True,
+        help="The main identifier of the currency.")
+    symbol = fields.Char('Symbol', size=10, required=True,
+        help="The symbol used for currency formating.")
+    code = fields.Char('Code', size=3, required=True,
+        help="The 3 chars ISO currency code.")
+    numeric_code = fields.Char('Numeric Code', size=3,
+        help="The 3 digits ISO currency code.")
+    rate = fields.Function(fields.Numeric('Current rate', digits=(12, 6)),
+        'get_rate')
+    rates = fields.One2Many('currency.currency.rate', 'currency', 'Rates',
+        help="Add floating exchange rates for the currency.")
+    rounding = fields.Numeric('Rounding factor', required=True,
+        digits=(12, Eval('digits', 6)), depends=['digits'],
+        help="The minimum amount which can be represented in this currency.")
+    digits = fields.Integer("Digits", required=True,
+        help="The number of digits to display after the decimal separator.")
 
     @classmethod
     def __setup__(cls):
@@ -45,16 +46,22 @@ class Currency(ModelSQL, ModelView):
         cls._error_messages.update({
                 'no_rate': ('No rate found for currency "%(currency)s" on '
                     '"%(date)s"'),
-                'invalid_mon_grouping': ('Invalid grouping "%(grouping)s" on '
-                    ' currency "%(currency)s".'),
                 })
         cls.__rpc__.update({
-                'compute': RPC(),
+                'compute': RPC(instantiate=slice(0, 3, 2)),
                 })
 
-    @staticmethod
-    def default_active():
-        return True
+    @classmethod
+    def __register__(cls, module_name):
+        super(Currency, cls).__register__(module_name)
+
+        table_h = cls.__table_handler__(module_name)
+
+        # Migration from 4.6: removal of monetary
+        for col in [
+                'mon_grouping', 'mon_decimal_point',
+                'p_sign_posn', 'n_sign_posn']:
+            table_h.not_null_action(col, 'remove')
 
     @staticmethod
     def default_rounding():
@@ -64,97 +71,38 @@ class Currency(ModelSQL, ModelView):
     def default_digits():
         return 2
 
-    @staticmethod
-    def default_mon_grouping():
-        return '[]'
-
-    @staticmethod
-    def default_mon_thousands_sep():
-        return ','
-
-    @staticmethod
-    def default_mon_decimal_point():
-        return '.'
-
-    @staticmethod
-    def default_p_sign_posn():
-        return 1
-
-    @staticmethod
-    def default_n_sign_posn():
-        return 1
-
-    @staticmethod
-    def default_negative_sign():
-        return '-'
-
-    @staticmethod
-    def default_positive_sign():
-        return ''
-
-    @staticmethod
-    def default_p_cs_precedes():
-        return True
-
-    @staticmethod
-    def default_n_cs_precedes():
-        return True
-
-    @staticmethod
-    def default_p_sep_by_space():
-        return False
-
-    @staticmethod
-    def default_n_sep_by_space():
-        return False
-
-    @classmethod
-    def validate(cls, currencies):
-        super(Currency, cls).validate(currencies)
-        for currency in currencies:
-            currency.check_mon_grouping()
-
-    def check_mon_grouping(self):
-        '''
-        Check if mon_grouping is list of numbers
-        '''
-        try:
-            grouping = safe_eval(self.mon_grouping)
-            for i in grouping:
-                if not isinstance(i, int):
-                    raise ValueError
-        except Exception:
-            self.raise_user_error('invalid_mon_grouping', {
-                    'grouping': self.mon_grouping,
-                    'currency': self.rec_name,
-                    })
-
     @classmethod
     def check_xml_record(cls, records, values):
         return True
 
+    @classmethod
+    def search_global(cls, text):
+        for record, rec_name, icon in super(Currency, cls).search_global(text):
+            icon = icon or 'tryton-currency'
+            yield record, rec_name, icon
+
     @classmethod
     def search_rec_name(cls, name, clause):
         currencies = None
         field = None
         for field in ('code', 'numeric_code'):
-            currencies = cls.search([(field,) + clause[1:]], limit=1)
+            currencies = cls.search([(field,) + tuple(clause[1:])], limit=1)
             if currencies:
                 break
         if currencies:
-            return [(field,) + clause[1:]]
-        return [(cls._rec_name,) + clause[1:]]
+            return [(field,) + tuple(clause[1:])]
+        return [(cls._rec_name,) + tuple(clause[1:])]
 
+    @fields.depends('rates')
     def on_change_with_rate(self):
         now = datetime.date.today()
         closer = datetime.date.min
         res = Decimal('0.0')
         for rate in self.rates or []:
-            if 'date' not in rate or 'rate' not in rate:
-                continue
-            if rate['date'] <= now and rate['date'] > closer:
-                res = rate['rate']
-                closer = rate['date']
+            date = getattr(rate, 'date', None) or now
+            if date <= now and date > closer:
+                res = rate.rate
+                closer = date
         return res
 
     @staticmethod
@@ -175,7 +123,7 @@ class Currency(ModelSQL, ModelView):
             if rates:
                 res[currency.id] = rates[0].id
             else:
-                res[currency.id] = 0.0
+                res[currency.id] = 0
         rate_ids = [x for x in res.values() if x]
         rates = Rate.browse(rate_ids)
         id2rate = {}
@@ -187,16 +135,7 @@ class Currency(ModelSQL, ModelView):
         return res
 
     def round(self, amount, rounding=ROUND_HALF_EVEN):
-        '''
-        Round the amount depending of the currency
-
-        :param cursor: the database cursor
-        :param user: the user id
-        :param currency: a BrowseRecord of currency.currency
-        :param amout: a Decimal
-        :param rounding: the rounding option
-        :return: a Decimal
-        '''
+        'Round the amount depending of the currency'
         return (amount / self.rounding).quantize(Decimal('1.'),
                 rounding=rounding) * self.rounding
 
@@ -213,8 +152,8 @@ class Currency(ModelSQL, ModelView):
         '''
         Date = Pool().get('ir.date')
         Lang = Pool().get('ir.lang')
-        from_currency = cls(from_currency.id)
-        to_currency = cls(to_currency.id)
+        from_currency = cls(int(from_currency))
+        to_currency = cls(int(to_currency))
 
         if to_currency == from_currency:
             if round:
@@ -241,23 +180,60 @@ class Currency(ModelSQL, ModelView):
         else:
             return amount * to_currency.rate / from_currency.rate
 
+    @classmethod
+    def currency_rate_sql(cls):
+        "Return a SQL query with currency, rate, start_date and end_date"
+        pool = Pool()
+        Rate = pool.get('currency.currency.rate')
+        transaction = Transaction()
+        database = transaction.database
+
+        rate = Rate.__table__()
+        if database.has_window_functions():
+            window = Window(
+                [rate.currency],
+                order_by=[rate.date.asc],
+                frame='ROWS', start=0, end=1)
+            # Use NthValue instead of LastValue to get NULL for the last row
+            end_date = NthValue(rate.date, 2, window=window)
+        else:
+            next_rate = Rate.__table__()
+            end_date = next_rate.select(
+                next_rate.date,
+                where=(next_rate.currency == rate.currency)
+                & (next_rate.date > rate.date),
+                order_by=[next_rate.date.asc],
+                limit=1)
+
+        query = (rate
+            .select(
+                rate.currency.as_('currency'),
+                rate.rate.as_('rate'),
+                rate.date.as_('start_date'),
+                end_date.as_('end_date'),
+                ))
+        return query
+
 
 class Rate(ModelSQL, ModelView):
     'Rate'
     __name__ = 'currency.currency.rate'
-    _rec_name = 'date'
-    date = fields.Date('Date', required=True, select=True)
-    rate = fields.Numeric('Rate', digits=(12, 6), required=1)
+    date = fields.Date('Date', required=True, select=True,
+        help="From when the rate applies.")
+    rate = fields.Numeric('Rate', digits=(12, 6), required=1,
+        help="The floating exchange rate used to convert the currency.")
     currency = fields.Many2One('currency.currency', 'Currency',
-            ondelete='CASCADE',)
+            ondelete='CASCADE',
+        help="The currency on which the rate applies.")
 
     @classmethod
     def __setup__(cls):
         super(Rate, cls).__setup__()
+        t = cls.__table__()
         cls._sql_constraints = [
-            ('date_currency_uniq', 'UNIQUE(date, currency)',
+            ('date_currency_uniq', Unique(t, t.date, t.currency),
                 'A currency can only have one rate by date.'),
-            ('check_currency_rate', 'CHECK(rate >= 0)',
+            ('check_currency_rate', Check(t, t.rate >= 0),
                 'The currency rate must greater than or equal to 0'),
             ]
         cls._order.insert(0, ('date', 'DESC'))
@@ -268,5 +244,8 @@ class Rate(ModelSQL, ModelView):
         return Date.today()
 
     @classmethod
-    def check_xml_record(self, records, values):
+    def check_xml_record(cls, records, values):
         return True
+
+    def get_rec_name(self, name):
+        return str(self.date)
